/*  дерево символов, слов и фраз
вербальная иерархия распознавателей

*/

package word_sensor

import (
	"BOT/lib"
	"sort"
	"strconv"
	"strings"
)

/////////////////////////////////////////////
// Первый уровень дерева - символы. Заглавных не может быть в воспринимаемом
var symbolsArr = []string{" ","а","б","в","г","д","е","ё","ж","з","и","й","к","л","м","н","о","п","р","с","т","у","ф","х","ц","ч","ш","щ","ъ","ы","ь","э","ю","я","a","b","c","d","e","f","g","h","i","j","k","l","m","n","o","p","q","r","s","t","u","v","w","x","y","z","!","?","@","#","$","%","^","&","*","(",")","+","=","-","1","2","3","4","5","6","7","8","9","0","[","]","{","}","<",">",".",",","/"}
func IsSymbol(w string)(bool){
	for i := 0; i < len(symbolsArr); i++ {
		if w == symbolsArr[i]{
			return true
		}
	}
	return false
}
/////////////////////////////////////////




////////////////////////////////////////////////////////////////////
type WordTree struct {
	ID int // id слова-фразы
	Word string
	Children []WordTree // дочерние узлы (ветвление) НЕ АДРЕСА, А РЕАЛЬНЫЕ ОБЪЕКТЫ
	ParentID int     // ID родителя
	ParentNode *WordTree  // адрес родителя
}

var VernikeWordTree WordTree
var WordTreeFromID=make(map[int]*WordTree)
var WordTreeFromStr=make(map[string][]*WordTree)// массив узлов с такой частью слова
/////////////// для обеспечения уникальности узлов:
type WordUnicum struct {
	ID int
	word string
}
var WordUnicumIdStr=make(map[WordUnicum]int)// для каждого сочетания  выдается ID узла
//////////////////////////////////////////////////
var lastIDwordTree=0
///////////////////////////////////////////////////////////////////////

// подошла очередь инициализации
func afterLoadTempArr(){
	loadWordTree()
	/*
	   	//а я нехочу
	   	new:=createNewNodeWordTree(WordTreeFromID[2],0," я")
	   	new=createNewNodeWordTree(new,0," нехочу")
	   	createNewNodeWordTree(new,0," нечего знать")
	   	saveWordTree()



	   //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	   	SetNewWordTreeNode("актуальными") // а я нехочу такое дело
	   //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	   	saveWordTree()
	*/


//	updateWordTreeFromTempArr(2,2)


	afetrInitWordTree()
}

// создать первый, символьный уровень дерева
func initWordTree(vt *WordTree){
	for i := 0; i < len(symbolsArr); i++ {
		createNewNodeWordTree(vt,0,symbolsArr[i])
	}

	return
}
/////////////////////////////////

func createNewNodeWordTree(parent *WordTree,id int,word string)(*WordTree){
	if parent==nil{
		return nil
	}
	/*
	// не позволять дублировать узлы
	unic:=WordUnicumIdStr[WordUnicum{parent.ID,word}]
	if unic!=0 {// уже есть такой узел
		old:=WordTreeFromID[unic] // WordTreeFromID еще актуальный для всех уже созданных, раз не создано нового
		return old //nil
	}
	*/

	if id==0{
		lastIDwordTree++;
		id=lastIDwordTree
	}else{
		//		newW.ID=id
		if lastIDwordTree<id{
			lastIDwordTree=id
		}
	}



// не позволять дублировать узлы
	unic:=WordUnicumIdStr[WordUnicum{id,word}]
	if unic!=0 {// уже есть такой узел
		old:=WordTreeFromID[unic] // WordTreeFromID еще актуальный для всех уже созданных, раз не создано нового
		return old //nil
	}

	var newW WordTree
	newW.ID = id
	newW.ParentID = parent.ID
	newW.ParentNode = parent
	newW.Word = word

	parent.Children = append(parent.Children, newW)
	// четко находим новый вставленный член (а не &parent.Children[count-1])
	var new *WordTree
	for i := 0; i < len(parent.Children); i++ {
		if parent.Children[i].ID == newW.ID {
			new = &parent.Children[i]
		}
	}

	//WordTreeFromID[new.ID]=new
	// т.к. append меняет длину массива, перетусовывая адреса, то нужно:
	updateWordTreeFromID()// здесь потому, что при загрузке из файла нужно на лету получать адреса

	return new
}
// корректируем адреса всех узлов
func updateWordTreeFromID(){
	updatingWordTreeFromID(&VernikeWordTree)
}
// проход всего дерева
func updatingWordTreeFromID(wt *WordTree){
	if wt.ID>0 {
		wt.ParentNode=WordTreeFromID[wt.ParentID] // wt.ParentNode адрес меняется из=за corretsParent(,
		WordTreeFromID[wt.ID] = wt
		WordTreeFromStr[wt.Word] = append(WordTreeFromStr[wt.Word],wt)
		WordUnicumIdStr[WordUnicum{wt.ParentID,wt.Word}] = wt.ID
	}
	if wt.Children == nil{// конец ветки
		return
	}
	for i := 0; i < len(wt.Children); i++ {
		updatingWordTreeFromID(&wt.Children[i])
	}
}
///////////////////////////////////////////////////////////


func loadWordTree(){

	initWordTree(&VernikeWordTree)
	strArr,_:=lib.ReadLines(lib.GetMainPathExeFile()+"/memory_reflex/word_tree.txt")
	cunt:=len(strArr)
	//просто проход по всем строкам файла подряд так что сначала идут дочки, потом - их родители
	for n := 0; n < cunt; n++ {
		if len(strArr[n])<2{
			return
		}
		p:=strings.Split(strArr[n], "|#|")
		word:=p[1]
		idP:=strings.Split(p[0], "|")
		id,_:=strconv.Atoi(idP[0])
		parentID,_:=strconv.Atoi(idP[1])
		// новый узел с каждой строкой из файла
		createNewNodeWordTree(WordTreeFromID[parentID],id,word)
	}
	return
}
////////////////////////////////////////////
func saveWordTree(){
	var out=""
	// записываем, начиная со второго уровня (уровень символов не пишем)
	cnt:=len(VernikeWordTree.Children)
	for n := 0; n < cnt; n++ {
		out+=getWtreeNode(&VernikeWordTree.Children[n])
	}
	lib.WriteFileContent(lib.GetMainPathExeFile()+"/memory_reflex/word_tree.txt",out)
	return
}
func getWtreeNode(wt *WordTree)(string){
	var out=""
	if wt.ParentID>0 {
		out += strconv.Itoa(wt.ID) + "|"
		out += strconv.Itoa(wt.ParentID) + "|#|"
		out += wt.Word + "\r\n"
	}
	if(wt.Children==nil){// конец
		return out
	}
	for n := 0; n < len(wt.Children); n++ {
		out+=getWtreeNode(&wt.Children[n])
	}
	return out
}
//////////////////////////////////////////////////////////////////////////



/////////////////////////////////////////////////////////////////////////

/* переносим в дерево слов достаточно повторяющиеся из tempArr
 limitWord - число повторений отдельных слов, с которых начинается пеернос.
limitFraze - число повторений отдельных фраз (с несколькими словами), с которых начинается пеернос.
Для автозаливки - (4,6),
для сообщения с Пульта - (2,4)
*/
func updateWordTreeFromTempArr(limitWord int,limitFraze int){
	//Удалять использованные строки накопительного массива
var newTempFileStr="" // это будет записано после процесса в /memory_reflex/words_temp_arr.txt

var str[]string // для выбранных повторяющихся слов - чтобы потом отсортирвать
	// чтобы фразы могли использовать слова (минимизация размера дерева)
		// слово или фраза?

	// сначала прходим только слова, потом - только фразы,
	for k, v := range tempArr {
		sps:=strings.Split(k, " ")
		if len(sps)==1{// это слово
			if v >= limitFraze {
				str = append(str, k)
			}else{
				newTempFileStr+=strconv.Itoa(v)+"|#|"+k+"\r\n"
			}
		}
	}
	sort.Strings(str)  // по алфавиту, чтобы максимально облегчить последовательное разделение слов
	for i := 0; i < len(str); i++ {
		cur:=str[i]
		SetNewWordTreeNode(cur)

//		saveWordTree() // для пошагового контроля
//		if(i>1){}
	}

	// проход для фраз
	str=nil
	for k, v := range tempArr {
		sps:=strings.Split(k, " ")
		if len(sps)>1{// это фраза
			if v >= limitFraze {
				str = append(str, k)
			}else{
				newTempFileStr+=strconv.Itoa(v)+"|#|"+k+"\r\n"
			}
		}
	}
	sort.Strings(str)  // по алфавиту, чтобы максимально облегчить последовательное разделение слов
	for i := 0; i < len(str); i++ {
		cur:=str[i]
		SetNewWordTreeNode(cur)

		//		saveWordTree() // для пошагового контроля
		//		if(i>1){}
	}

	saveWordTree()

	// Удаление использованных строк накопительного массива: запись только незатронутых
	lib.WriteFileContent(lib.GetMainPathExeFile()+"/memory_reflex/words_temp_arr.txt",newTempFileStr)
}
/////////////////////////////////////////////////

// вставить новое слово-фразу в дерево:
//найти подходящий узел и если еще нет - вставить новый.
var level=0
func SetNewWordTreeNode(word string)(*WordTree){
	if len(word)==0{
		return nil
	}
	var last *WordTree
	// первый символ
	r := []rune(word)
	// основа дерева
	cnt:=len(VernikeWordTree.Children)
	for n := 0; n < cnt; n++ {
		rt := []rune(VernikeWordTree.Children[n].Word)
		if r[0]==rt[0]{
			cldrn:=VernikeWordTree.Children[n].Children;
			cnt:=len(cldrn)
			if cnt==0{// нет дочек - просто создать
				ost:=r[1:]
				createNewNodeWordTree(&VernikeWordTree.Children[n], 0, string(ost))
				break
			}
			for n := 0; n < cnt; n++ {
				level=1
				last = findWordTreeNode(level, r[1:], &cldrn[n])
				if last==nil{// найдено, больше не смотреть
					return nil
				}
			}
		}
	}
	return last
}
///////////////////
func findWordTreeNode(level int,word []rune,wt *WordTree)(*WordTree){
/*
	if string(word)=="втоматизма"{
		if len(word)==0{}
	}
*/
	if len(word)==0{// конец, слово распознано
		return wt
	}

	rt := []rune(wt.Word)
	var ost=word
	//var rOst=word
	var i0=0
	for i := 0; i < len(rt) && i < len(word); i++ {
		i0=i+1
		if len(word)>i{
			ost = word[i0:]
		}
		//rOst=word[i:len(rt)]
		//ost=word[i:len(rt)]
		if rt[i] != word[i] { //  перестало совпадать
			if i == 0 { // не совпадает полностью с первой же буквы
				parent:=WordTreeFromID[wt.ParentNode.ID]
				cnt:=len(parent.Children)
//или создать новый узел или продолжить в подходящем узле parent.Children[n]
				for n := 0; n < cnt; n++ {
					rt = []rune(parent.Children[n].Word)
					//если у детей parent есть узел, начинающийся на первую букву word
					if string(word[0])==string(rt[0]){
						//продолжить поиск в этом узле для разбивки слова узла или окончания поиска
						last := findWordTreeNode(level,word, &parent.Children[n])
						return last
					}
				}
//если у детей parent нет узлов, начинающихся на первую букву word, то открыть такой узел со всем word
				createNewNodeWordTree(parent, 0, string(word))
				return nil// все найдено, больше дочек не проходить

			} else { // часть совпадает, значит нужно разделть слово узла на два
				so := string(rt[i:])
				wo := string(word[i:])
				wt.Word = string(rt[:i])
				//old:=wt.Children[n].ID

				new := createNewNodeWordTree(wt, 0, so)
				corretsParent(wt, new)
				// родителем остается прежний wt.Children[n]
				//new=createNewNodeWordTree(wt, 0, wo); return new
				createNewNodeWordTree(wt, 0, wo); return nil
			}
		}


	}//for i := 0; i < len(rt); i++ {
	//}// if level>0

	// все совпало с узлом до конца, но еще есть остаточк слова



	if(wt.Children==nil ){// конец

		if len(ost)==0{// закончен проход слова, поиск завершен
			return nil
		}

		if len(ost)>0 { // но еще есть фраза
			createNewNodeWordTree(WordTreeFromID[wt.ID], 0, string(ost))
			return nil // все найдено, больше дочек не проходить
		}
		/*
		if len(word)>0{// но еще есть фраза -
			// дорастить в дочке узла wt
			//	new:=createNewNodeWordTree(wt.ParentNode, 0, string(word))  //- МОЖЕТ НЕ ПРОКАТЫТИТЬ
			createNewNodeWordTree(WordTreeFromID[wt.ParentNode.ID], 0, string(word))
			return nil // все найдено, больше дочек не проходить
		}
		 */
		return wt // найден узел
	}



	cnt:=len(wt.Children)
	for n := 0; n < cnt; n++ {
		level++
		last := findWordTreeNode(level,ost, &wt.Children[n])
		return last
	}
	return nil
}
//достаточно скорректирвовать у первого уровня дочек, у остсльных остаются их связы
func corretsParent(old *WordTree,new *WordTree){
// Все дочки кроме новой нужно перенести в новую, т.к. образом переренеся ветку дочек в ноьсозданную.
//Скорретирвоать все затронутые WordTreeFromID - вроде итак нормально так что не корректирую.
var childrens []WordTree // новый список дочек у old
	sldn:=old.Children
	for i := 0; i < len(sldn); i++ {
//только для sldn[i].ID == new.ID || (sldn[i].ParentID==old.ID)
		if (sldn[i].ID == new.ID) {//только что вставленную дочку не корретировать
			childrens=append(childrens,*new)
			continue
		}
// добавить дочки к new:
		sldn[i].ParentID=new.ID ;
		new.Children=append(new.Children,sldn[i])
//		sldn[i].ParentID=new.ID ;	sldn[i].ParentNode=new
	}
	// новый список дочек - с новой единственной дочкой второй отрезанной части слова
		old.Children = childrens
}
////////////////////////////////////////////////////////////

